from mcdreforged.plugin.plugin_event import LiteralEvent, PluginEvent, MCDRPluginEvents

__all__ = [
	# If you want to customize, inherit this
	'PluginEvent',

	# A simple event type
	'LiteralEvent',

	# Built-in MCDR events
	'MCDRPluginEvents',
]
