"""
Some custom exceptions in MCDR that might be useful in plugin dev
"""
from mcdreforged.utils.exception import IllegalCallError, IllegalStateError

__all__ = [
	'IllegalCallError', 'IllegalStateError'
]
